module ActiveResource::Hel::Extensions

  #
  # Module that gets mixed in the Resource Connection to handle Hel style
  # authentication
  #
  module Authentication
    extend ActiveSupport::Concern

    included do
      prepend PrependMethods
    end

    module PrependMethods
      def get(path, headers = {})
        with_authentication do | auth_headers |
          headers = headers.merge(auth_headers)
          super(path, headers)
        end
      end

      def delete(path, headers = {})
        with_authentication do | auth_headers |
          headers = headers.merge(auth_headers)
          super(path, headers)
        end
      end

      def patch(path, body = "", headers = {})
        with_authentication do | auth_headers |
          headers = headers.merge(auth_headers)
          super(path, body, headers)
        end
      end

      def put(path, body = "", headers = {})
        with_authentication do | auth_headers |
          headers = headers.merge(auth_headers)
          super(path, body, headers)
        end
      end

      def post(path, body = "", headers = {})
        with_authentication do | auth_headers |
          headers = headers.merge(auth_headers)
          super(path, body, headers)
        end
      end

      def head(path, headers = {})
        with_authentication do | auth_headers |
          headers = headers.merge(auth_headers)
          super(path, headers)
        end
      end
    end

    def with_authentication(&block)
      session.with_authentication do |headers|
        yield headers
      end
    end

    def authenticate!
      session.authenticate!
    end

    def check!
      res = get ActiveResource::Hel::Session::SESSION_CHECK_METHOD
      ActiveSupport::JSON.decode(res.body)['authenticated']
    end

    #
    # returns the current connection authentication context
    #
    def session
      self.current_session ||=  new_session(secret: hel_secret, identity: hel_identity)
    end

    #
    # returns a new authentication session for the given params
    #
    # mysession = connection.new_session(:identity => "foo", :secret => "bar")
    #
    # mysession = connection.new_session(:token => 'aa1e0c79df691159c06ce59b76e3972a')
    #
    def new_session(params = {})
      params = params.merge(:site => site)
      params = params.merge(:ssl_options => ssl_options)
      ActiveResource::Hel::Session.new(params)
    end

    #
    # executes the block in the context of an explicit session
    #
    # mysession = connection.new_session(:user => "admin", :pass => "s3kr337")
    # connection.with_session mysession do
    #   User.get(23)
    # end
    #
    # connection.with_session :token => 'aa1e0c79df691159c06ce59b76e3972a' do
    #   Identity.find(...)
    # end
    #
    def with_session(session, &block)
      if session.is_a?(Hash)
        session = new_session(session)
      elsif session.is_a? Ygg::Session
        session = new_session.initialize_from_data(session.attributes)
      end

      self.standby_session = current_session
      self.current_session = session
      res = block.call session

      self.current_session = standby_session
      self.standby_session = nil
      res
    end


    def current_session
      Thread.current[:ar_session]
    end

    def current_session=(session)
      Thread.current[:ar_session] = session
    end
  end
end
