module ActiveResource::Hel::Extensions

  #
  # Module that gets mixed in the Resource Connection to handle Hel style
  # authentication
  #
  module Cache
    class NotModified < Exception
    end

    extend ActiveSupport::Concern

    included do
      prepend PrependMethods
    end

    attr_accessor :cached

    module PrependMethods
      def get(path, headers = {})
        with_cache(path, headers) do | cache_headers |
          super(path, cache_headers)
        end
      end

      def head(path, headers = {})
        with_cache(path, headers) do | cache_headers |
          super(path, cache_headers)
        end
      end
    end

    private

    def with_cache(path, headers = {}, &block)
      if self.cached
        do_request_with_cache(path, headers, &block)
      else
        do_request_without_cache(path, headers, &block)
      end
    end

    def do_request_without_cache(path, headers = {}, &block)
      yield(headers)
    end

    def do_request_with_cache(path, headers = {}, &block)
      cached_data = cache_read(path, headers) || {}
      response = cached_data[:response]
      if stale?(path, cached_data)
        cache_headers = headers.merge(headers_from_cache(cached_data))
        response = yield(cache_headers)
        raise NotModified.new if response.is_a?(Net::HTTPNotModified)
        cache_write(path, headers, response)
      else
        logger.info  "CACHE #{path}" if logger
      end
      response 
    rescue NotModified
      cache_touch(path, headers, cached_data)
      cached_data[:response]
    end

    def cache_read(path, headers)
      data = ActiveResource::Hel::Resource.cache.read(cache_key(path, headers))
      logger.debug "+++ Read response from cache: etag: #{data[:etag]} last-modified: #{data[:last_modified]}, ttl: #{data[:ttl]}" if logger && data
      data
    end

    def cache_write(path, req_headers, response)
      etag, last_modified, can_cache, ttl, must_revalidate = parse_cache_headers(response)
      if can_cache
        logger.debug "+++ Write response to cache: etag: #{etag} last-modified: #{last_modified || Time.new}, ttl: #{ttl}" if logger
        data = {
          cached_at: Time.new,
          etag: etag,
          last_modified: last_modified || Time.new,
          ttl: ttl,
          response: response,
          cache_version: cache_version(path)
        }
        ActiveResource::Hel::Resource.cache.write(cache_key(path, req_headers), data)
      end
    end

    def cache_touch(path, req_headers, data)
      data[:cached_at] = Time.new
      logger.debug "+++ Touching response cache" if logger
      ActiveResource::Hel::Resource.cache.write(cache_key(path, req_headers), data)
    end

    def parse_cache_headers(response)
      etag = response['etag']
      last_mod = response['last_modified']
      ctrl = response['cache-control']
      ctrl = Hash[ctrl.split(',').map(&:strip).map{|t| t = t.split('='); t << true; t[0..1] }] if ctrl
      can_cache = ctrl && !ctrl['no-store']
      ctrl ||= {}
      ttl = ((ctrl['max-age']) || 0).to_i
      ttl = 0 if ctrl['no-cache']
      logger.debug "+++ cache headers: etag: #{etag} last_modified: #{last_mod.inspect} can_cache: #{can_cache}, ttl: #{ttl}" if logger
      [etag, last_mod, can_cache, ttl]
    end

    def stale?(path, data)
      data ||= {}
      stale = false
      stale = true if data[:cached_at].blank?
      stale = true if !stale && data[:cache_version] && data[:cache_version] != cache_version(path)
      stale = true if !stale && data[:ttl] && data[:cached_at] && data[:cached_at] + data[:ttl].seconds < Time.new
      stale
    end

    def headers_from_cache(data)
      data ||= {}
      headers = {}
      if data[:etag] || data[:last_modified]
        headers['If-None-Match'] = data[:etag] if data[:etag].present? 
        headers['If-Modified-Since'] = data[:last_modified].utc.strftime('%a, %d %b %Y %T GMT') if data[:last_modified].present?
      end
      headers
    end

    def cache_key(path, headers)
      Digest::MD5.hexdigest("#{path}|#{format.mime_type}|#{headers['Authorization']}")
    end

    def cache_version(path)
      path = path.gsub(/\?.*/, '').split('/')
      keys = []
      path.inject([]) do | memo, part |
        memo << part
        keys << ActiveResource::Hel::Resource.cache.read("#{memo.join('/')}-cache-version")
        memo
      end
      keys.compact.max || 0      
    end
  end
end
