module ActiveResource::Hel
  module Filters
    extend ActiveSupport::Concern

    included do
      class << self
        prepend PrependClassMethods
      end
    end

    module PrependClassMethods
      def query_string(options)
        unless options.nil? || options.empty?
          options[:filter] = build_filter(options[:filter]).to_json if options[:filter].is_a?(Proc)
          options[:filter] = options[:filter].to_json if options[:filter].is_a?(Hash) || options[:filter].is_a?(ActiveResource::Hel::Filters::DSL)
        end
        super(options)
      end
    end

    module ClassMethods

      # Build yggdra query hash
      # example:
      # build_query do
      #   eq(field1: 'value1')
      #   self.or{
      #     self.and{
      #       gt(field2: value)
      #       lt(field3: value)
      #     }
      #   }
      # end
      # yelds a query like:
      # field1 = 'value1' OR field2 > value AND field3 < value
      #
      # operators:
      # eq          -> =
      # not_eq      -> <>
      # gt          -> >
      # gteq        -> >=
      # lt          -> <
      # lteq        -> <=
      # in          -> IN array
      # not_in      -> NOT IN array
      # like        -> LIKE (with the sql wildcards)
      # not_like    -> NOT LIKE (with the sql wildcards)
      # is_null     -> IS NULL (accepts array of fields)
      # is_not_null -> IS NOT NULL (accepts array of fields)
      # regexp      -> REGEXP ...
      # not_regexp  -> NOT_REGEXP ...
      #
      # or {}     -> join queries with OR
      # and {}    -> join queries with AND
      #
      # all operators are joined with AND except into a 'or' block
      # all operators except #and & #or accepts an hash
      # all key-value pairs are joined with AND, except inside a 'or' block
      #   eq(a: 'a', b: 'b', c: 'c') -> a = 'a' AND b = 'b' AND c = 'c'
      #   or{eq(a: 'a', b: 'b', c: 'c')} -> a = 'a' OR b = 'b' OR c = 'c'
      # 'and' & 'or' must be called as self.and & self.or (both are ruby keywords)
      #
      # filters are chainable
      # filter = build_filter
      # filter.build do
      #   xyx
      # end
      # filter.build do
      #   abc
      # end
      # filter.to_json
      #
      # finders that accepts params: {filter: ...}, also accepts an instance of builder object or a proc
      # Resource.find(:first, params: {filter: filter})
      # Resource.find(:first, params: {filter: ->{ self.or{eq(a: 'a'); eq(a: 'b')} }})

      def build_filter(proc = nil, &block)
        dsl = DSL.new
        dsl.build(&block) if proc.nil? && block_given?
        dsl.build(&proc) if proc.present?
        dsl
      end
    end

    class DSL
      OPERATORS = {
        eq: '=',
        not_eq: '<>',
        gt: '>',
        gteq: '>=',
        lt: '<',
        lteq: '<=',
        like: 'ILIKE',
        not_like: 'NOT ILIKE',
        in: 'IN',
        not_in: 'NOT IN',
        regexp: 'REGEXP',
        not_regexp: 'NOT REGEXP',
      }

      def initialize(default_join = AndQuery)
        @default_join = default_join
        @filter = nil
      end

      def build(&block)
        self.instance_exec(&block) if block_given?
        self
      end

      def as_json
        @filter.as_json
      end

      def to_json
        self.as_json.to_json
      end

      def or(&block)
        q = DSL.new(OrQuery).instance_eval &block
        q = OrQuery.new(@filter, q) if @filter
        @filter = q
        self
      end

      def and(&block)
        q = DSL.new(AndQuery).instance_eval &block
        q = AndQuery.new(@filter, q) if @filter
        @filter = q
        self
      end

      def is_null(conds)
        conds = conds.map do |k|
          IsNullQuery.new(k)
        end
        @filter = conds.inject(@filter) do | filter, query |
          if filter
            @default_join.new(filter, query)
          else
            query
          end
        end
        self
      end

      def is_not_null(conds)
        conds = conds.map do |k|
          IsNotNullQuery.new(k)
        end
        @filter = conds.inject(@filter) do | filter, query |
          if filter
            @default_join.new(filter, query)
          else
            query
          end
        end
        self
      end

      def method_missing(m, *args, &block)
        if OPERATORS[m].present?
          is_op(OPERATORS[m], args.first)
        else
          super
        end
      end

      protected

      def is_op(op, conds)
        conds = conds.map do |k, v|
          OperatorQuery.new(op, k,v)
        end
        @filter = conds.inject(@filter) do | filter, query |
          if filter
            @default_join.new(filter, query)
          else
            query
          end
        end
        self
      end

      class QueryElement
        class_attribute :operator
        self.operator = nil

        def initialize(a, b)
          @a = a
          @b = b
        end

        def as_json
          {a: {field: @a}, o: operator, b: @b}
        end
      end

      class AndQuery < QueryElement
        self.operator = 'AND'
        def as_json
          {a: @a.try(:as_json) || @a, o: operator, b: @b.try(:as_json) || @b}
        end
      end
      class OrQuery < QueryElement
        self.operator = 'OR'
        def as_json
          {a: @a.try(:as_json) || @a, o: operator, b: @b.try(:as_json) || @b}
        end
      end
      class OperatorQuery < QueryElement
        def initialize(op, a, b)
          self.operator = op
          super(a,b)
        end
      end
      class IsNullQuery < QueryElement
        self.operator = 'IS NULL'
        def initialize(a)
          @a = a
        end
        def as_json
          {a: {field: @a}, o: operator}
        end
      end
      class IsNotNullQuery < IsNullQuery
        self.operator = 'IS NOT NULL'
      end
    end

  end
end