module ActiveResource::Hel

  #
  # Extends ActiveResource::Base to provide a Hel resource of the appropriate format and
  # authentication
  #
  class Resource < ActiveResource::Base
    DEFAULT_CACHE = ActiveSupport::Cache::MemoryStore.new

    include ActiveResource::Hel::Scopes
    include ActiveResource::Hel::DefaultScopes
    include ActiveResource::Hel::RemotePagination
    include ActiveResource::Hel::RemoteSchema
    include ActiveResource::Hel::Validations
    include ActiveResource::Hel::Cached
    include ActiveResource::Hel::Filters
    include ActiveResource::Hel::Rails if defined?('Rails')

    class_attribute :cache

    self.connection_class = ActiveResource::Hel::Connection
    self.cache = DEFAULT_CACHE

    #
    # enhanced support for custom REST methods
    # they work like AR counterparts, but provide
    # - encoding of request objects
    # - decoding of answers
    # - they don't show AR behaviour of deleting root element in answers
    # - Hel authentication
    # - appropriate URIs
    #
    class << self
      include ThreadsafeAttributes
      threadsafe_attribute :_hel_identity, :_hel_secret

      #
      # trues to guess the type of embedded object instead of just returning hashes
      #
      def guess_embedded_types
        if defined?(@guess_embedded_types)
          @guess_embedded_types
        elsif superclass != ActiveResource::Base && (superclass.guess_embedded_types != nil)
          superclass.guess_embedded_types
        else
          @guess_embedded_types = false
        end
      end

      def guess_embedded_types=(guess)
        @guess_embedded_types = guess
      end

      #
      # encodes requests
      #
      def resource_codec(options, body, &block)
        body ||= {}

        encoded_body  = (body.is_a?(String) ? body : body.to_json)
        res = block.call(encoded_body)
        format.plain_decode(res.body)
      end

      #
      # custom REST class/collection methods
      #
      def get(custom_method_name, options = {})
        format.plain_decode(connection.get(custom_method_collection_url(custom_method_name, options), headers).body)
      end

      def post(custom_method_name, options = {}, body = '')
        resource_codec(options, body) do |encoded_body|
          super(custom_method_name, options, encoded_body)
        end
      end

      def put(custom_method_name, options = {}, body = '')
        resource_codec(options, body) do |encoded_body|
          super(custom_method_name, options, encoded_body)
        end
      end

      def delete(custom_method_name, options = {})
        res = super(custom_method_name, options)
        format.plain_decode(res.body)
      end

      def hel_identity
        # Not using superclass_delegating_reader. See +site+ for explanation
        if _hel_identity_defined?
          _hel_identity
        elsif superclass != Object && superclass.try(&:hel_identity)
          superclass.hel_identity.dup.freeze
        end
      end

      def hel_identity=(id)
        self._connection = nil
        self._hel_identity = id
      end

      def hel_secret
        # Not using superclass_delegating_reader. See +site+ for explanation
        if _hel_secret_defined?
          _hel_secret
        elsif superclass != Object && superclass.try(&:hel_secret)
          superclass.hel_secret.dup.freeze
        end
      end

      def hel_secret=(sec)
        self._connection = nil
        self._hel_secret = sec
      end

      def connection(refresh = false)
        if _connection_defined? || superclass == Object
          self._connection = connection_class.new(
            site, format, logger: logger
          ) if refresh || _connection.nil?
          _connection.proxy = proxy if proxy
          _connection.user = user if user
          _connection.password = password if password
          _connection.auth_type = auth_type if auth_type
          _connection.timeout = timeout if timeout
          _connection.open_timeout = open_timeout if open_timeout
          _connection.read_timeout = read_timeout if read_timeout
          _connection.ssl_options = ssl_options if ssl_options
          _connection.hel_identity = hel_identity if hel_identity
          _connection.hel_secret = hel_secret if hel_secret
          _connection.cached = cached
          _connection
        else
          superclass.connection
        end
      end

      #
      # Override some specific AR::Base method which we want to have slighly different
      # behaviours
      #

      def prefix(options={})
        default = site.path
        default << '/' unless default[-1..-1] == '/'
        default += "#{namespace}/" if namespace.present?

        # generate the actual method based on the current site path
        self.prefix = default
        prefix(options)
      end

      def namespace
        ActiveSupport::Inflector.deconstantize(self.name).underscore
      end

      def element_path(id, prefix_options = {}, query_options = nil)
        check_prefix_options(prefix_options)
        prefix_options, query_options = split_options(prefix_options) if query_options.nil?
        "#{prefix(prefix_options)}#{collection_name}/#{URI.parser.escape id.to_s}#{query_string(query_options)}"
      end

      def new_element_path(prefix_options = {})
        "#{prefix(prefix_options)}#{collection_name}/new"
      end

      def collection_path(prefix_options = {}, query_options = nil)
        check_prefix_options(prefix_options)
        prefix_options, query_options = split_options(prefix_options) if query_options.nil?
        "#{prefix(prefix_options)}#{collection_name}#{query_string(query_options)}"
      end

      def custom_method_collection_url(method_name, options = {})
        prefix_options, query_options = split_options(options)
        "#{prefix(prefix_options)}#{collection_name}/#{method_name}#{query_string(query_options)}"
      end


      def with_session(session, &block)
        connection.with_session(session, &block)
      end

      def find(*arguments)
        ensure_remote_schema
        super
      end

      private

      # Find every resource
      def find_every(options)
        begin
          case from = options[:from]
          when Symbol
            response = connection.get(custom_method_collection_url(from, options[:params]))
            instantiate_collection(format.decode(response.body) || [], response, options[:params])
          when String
            path = "#{from}#{query_string(options[:params])}"
            response = connection.get(path, headers)
            instantiate_collection(format.decode(response.body) || [], response, options[:params])
          else
            prefix_options, query_options = split_options(options[:params])
            path = collection_path(prefix_options, query_options)
            response = connection.get(path, headers)
            instantiate_collection( (format.decode(response.body) || []), response, query_options, prefix_options )
          end
        rescue ActiveResource::ResourceNotFound
          # Swallowing ResourceNotFound exceptions and return nil - as per
          # ActiveRecord.
          nil
        end
      end

      if ActiveResource::VERSION::MAJOR >= 4
        def instantiate_collection(collection, response, original_params = {}, prefix_options = {})
          total_count = response.to_hash['x-total-resources-count'].try(:first).to_i
          ActiveResource::Hel::Collection.new(super(collection, original_params, prefix_options), {
            :total_count => total_count,
            :offset => (response.to_hash['x-resources-offset'].try(:first) || 0).to_i,
            :limit => (response.to_hash['x-resources-limit'].try(:first) || total_count).to_i
          })

        end
      else
        def instantiate_collection(collection, response, prefix_options = {})
          total_count = response.to_hash['x-total-resources-count'].try(:first).to_i
          ActiveResource::Hel::Collection.new(super(collection, prefix_options), {
            :total_count => total_count,
            :offset => (response.to_hash['x-resources-offset'].try(:first) || 0).to_i,
            :limit => (response.to_hash['x-resources-limit'].try(:first) || total_count).to_i
          })

        end
      end
    end


    def load_collection(key, value, persisted)
      resource = nil
      embedded = false

      value = value.map do |attrs|
        if attrs.is_a?(Hash)
          load_object(key, attrs, persisted, resource, true, embedded)
        else
          attrs.duplicable? ? attrs.dup : attrs
        end
      end

      if schema && schema[key]
        case schema[key]
          when 'embedded_collection'
            klass = reflections[key.to_s].try(:klass)
            value = ActiveResource::Hel::EmbeddedCollection.new(value, :model => klass,
                                                                :parent => self, :attribute => key)
            embedded = true

          when 'collection'
            klass = reflections[key.to_s].try(:klass)
            value = ActiveResource::Hel::Collection.new(value, :model => klass, :parent => self,
                                                        :attribute => key)
          else


        end
      end
      value
    end

    def load_object(key, value, persisted, resource_class = nil, collection = nil, embedded=nil)
      if (value.has_key?(:_type) || value.has_key?('_type')) &&
          (type = value.delete(:_type) { value.delete('_type') })
        resource_class ||= find_or_create_resource_by_name(type, embedded)
        resource_class.new(value, persisted)
      elsif self.class.guess_embedded_types
        resource_class ||= (collection ? find_or_create_resource_for_collection(key)
              : find_or_create_resource_for(key))
        resource_class.new(value, persisted)
      else
        value.dup
      end
    end

    def load(attributes, remove_root = false, persisted = false)
      raise ArgumentError, "expected an attributes Hash, got #{attributes.inspect}" unless attributes.is_a?(Hash)
      @prefix_options, attributes = split_options(attributes)

      self.class.ensure_remote_schema

      if attributes.keys.size == 1
        remove_root = self.class.element_name == attributes.keys.first.to_s
      end

      attributes = ActiveResource::Formats.remove_root(attributes) if remove_root

      attributes.each do |key, value|
        @attributes[key.to_s] =
          case value
            when Array, Collection
              load_collection(key, value, persisted)
            when Hash
              load_object(key, value, persisted)
            else
              value.duplicable? ? value.dup : value
          end
      end
      self
    end

    #
    # custom REST instance/item methods
    #
    def get(method_name, options = {})
      self.class.format.plain_decode(connection.get(custom_method_element_url(method_name, options), self.class.headers).body)
    end

    def post(custom_method_name, options = {}, body = '')
      self.class.resource_codec(options, body) do |encoded_body|
        super(custom_method_name, options, encoded_body)
      end
    end

    def put(custom_method_name, options = {}, body = '')
      self.class.resource_codec(options, body) do |encoded_body|
        super(custom_method_name, options, encoded_body)
      end
    end

    def delete(custom_method_name, options = {})
      res = super(custom_method_name, options)
      format.plain_decode(res.body)
    end

    #
    # returns an hash of the listed attributes
    #
    def filtered_attributes(*attrs)
      attributes.inject_hash(HashWithIndifferentAccess.new) { |acc, (n,v)| acc[n] = v if attrs.include?(n.to_sym) }
    end


    def attribute(name)
      attributes[name]
    end

    def set_attribute(name, value)
      self.send("#{name}=".to_sym, value)
    end

    def update_attributes(attributes)
      attributes.keys.each do |k|
        attribute_will_change! k
      end
      super
    end

    def assign_attribute(name, value)
      self.send("#{name}=".to_sym, value)
    end

    def assign_attributes(attributes)
      attributes.keys.each do |k|
        attribute_will_change! k
      end
      load(attributes)
    end

    def custom_save(method, options = {})
      perform_validation = options[:validate] != false
      options.merge!(:custom_method => method, :skip_remote_validation => true)

      # clear the remote validations so they don't interfere with the local
      # ones. Otherwise we get an endless loop and can never change the
      # fields so as to make the resource valid.
      @remote_errors = nil
      if perform_validation && valid?(options) || !perform_validation
        new? ? create(options) : update(options)
        true
      else
        false
      end
    rescue ActiveResource::ResourceInvalid => error
      # cache the remote errors because every call to <tt>valid?</tt> clears
      # all errors. We must keep a copy to add these back after local
      # validations.
      @remote_errors = error
      load_remote_errors(@remote_errors, true)
      false
    end

    def custom_save!(method, options = {})
      custom_save(method, options) || raise(ActiveResource::ResourceInvalid.new(self))
    end

    def read_attribute_for_serialization(n, options = {})

      case schema[n]
      when 'embedded_collection'

        collection = attributes[n]

        # changed elements
        res = collection.select { |x| !options[:dirty] || x.changed? }.inject([]) do |a, member|
          a << member.serializable_hash(options)
          a
        end

        collection.select { |x| x.new? }.each do |member|
          res.push(member.serializable_hash(options.reject{|k,v| k = :dirty }))
        end

        collection.destroyed.each do |e|
          res.push(:id => e, :_destroy => true)
        end
        res
      else
        attributes[n]
      end
    end

    def serializable_hash(options = nil)
      options ||= {}

      attribute_names = attributes.keys
      if only = options[:only]
        attribute_names &= Array(only).map(&:to_s)
      elsif except = options[:except]
        attribute_names -= Array(except).map(&:to_s)
      elsif options[:dirty]
        attribute_names &= Array(self.changed.append(:id)).map(&:to_s)
      end

      hash = {}
      attribute_names.each { |n| hash[n] = read_attribute_for_serialization(n, options) }

      Array(options[:methods]).each { |m| hash[m.to_s] = send(m) if respond_to?(m) }

      serializable_add_includes(options) do |association, records, opts|
        hash[association.to_s] = if records.respond_to?(:to_ary)
                                   records.to_ary.map { |a| a.serializable_hash(opts) }
                                 else
                                   records.serializable_hash(opts)
                                 end
      end

      hash
    end

    private


=begin
    def encode(options={})
      if options.delete(:for_update)
        #opts = options.merge(:only => self.changed.append(:id))

      end
      send("to_#{self.class.format.extension}", opts)
    end
=end

    #
    # adds dirty tracking to the default method
    #
    def update(options = {})
      additionals_headers = options[:only_validate] ? { 'X-Validate-Only' => 'true' } : {}
      additionals_headers.merge!(options[:additional_headers]) if options[:additional_headers]

      path = options[:custom_method] ? custom_method_element_url(options[:custom_method], prefix_options) : element_path(prefix_options)

      @previously_changed = changes
      run_callbacks :update do
        connection.put(path, encode(:dirty => true), self.class.headers.merge(additionals_headers)).tap do |response|
          unless options[:only_validate]
            changed_attributes.clear
            clear_embedded_changes
          end

          load_attributes_from_response(response)
        end
      end
    end

    # Create (i.e., \save to the remote service) the \new resource.
    # adds dirty tracking to the default method
    def create(options = {})
      additionals_headers = options[:only_validate] ? { 'X-Validate-Only' => 'true' } : {}
      additionals_headers.merge!(options[:additional_headers]) if options[:additional_headers]

      path = options[:custom_method] ? custom_method_new_element_url(options[:custom_method]) : collection_path

      run_callbacks :create do
        connection.post(path, encode, self.class.headers.merge(additionals_headers)).tap do |response|
          self.id = id_from_response(response)
          unless options[:only_validate]
            changed_attributes.clear
            clear_embedded_changes
          end
          load_attributes_from_response(response)
        end
      end
    end

    def clear_embedded_changes
      embedded_collections.each {|c| c.clear_embedded_changes }
    end

    def embedded_collections
      attributes.values.select {|k,v| v.is_a? EmbeddedCollection }
    end

    #
    # adds dirty tracking to the default method
    #
    def method_missing(method_symbol, *arguments) #:nodoc:
       method_name = method_symbol.to_s

       if method_name =~ /(=|\?)$/
         case $1
         when "="
           attribute_will_change! $`
           attributes[$`] = arguments.first
         when "?"
           attributes[$`]
         end
       else
         return attributes[method_name] if attributes.include?(method_name)
         # not set right now but we know about it
         return nil if known_attributes.include?(method_name)
         super
       end
    end

    def custom_method_element_url(method_name, options = {})
      "#{self.class.prefix(prefix_options)}#{self.class.collection_name}/#{id}/#{method_name}#{self.class.__send__(:query_string, options)}"
    end

    def custom_method_new_element_url(method_name, options = {})
      "#{self.class.prefix(prefix_options)}#{self.class.collection_name}/#{method_name}#{self.class.__send__(:query_string, options)}"
    end

    class << self
      def find_or_create_resource_by_name(name, embedded = false)
        if klass = class_exists?(name)
          klass
        else
          create_absolute_resource_for(name, embedded)
        end
      end

      def class_exists?(class_name)
        klass = class_from_string(class_name)
        klass.is_a?(Class) ? klass : false

        #ObjectSpace.each_object(Class) {|c| return true if c.to_s == class_name }
      rescue NameError
        false
      end

      def class_from_string(str)
        str.split('::').reject {|x| x.blank?}.inject(Object) do |mod, class_name|
          mod.const_get(class_name)
        end
      end

      def create_absolute_resource_for(resource_name, embedded = false)

        parent_class = (embedded ? ActiveResource::Hel::EmbeddedResource : ActiveResource::Hel::Resource)

        ns = find_or_create_namespace ActiveSupport::Inflector.deconstantize(resource_name)
        resource = ns.const_set(ActiveSupport::Inflector.demodulize(resource_name),
                                Class.new(ActiveResource::Hel::Resource))
        resource.prefix = prefix
        resource.site   = site
        resource
      end

      def find_or_create_namespace(ns)
        return Object if ns.blank?

        ns = ns.split('::') if ns.is_a? String
        name = ns.pop

        return Object if name.blank?

        parent = find_or_create_namespace(ns)
        parent.const_get(name)
      rescue NameError
        parent.const_set(name, Module.new)
        parent.const_get(name)
      end

      # Create and return a class definition for a resource inside the current resource
      def create_resource_for(resource_name)
        resource = const_set(resource_name, Class.new(ActiveResource::Hel::Resource))
        resource.prefix = prefix
        resource.site   = site
        resource
      end
    end

      delegate :find_or_create_resource_by_name, :create_absolute_resource_for, :create_resource_for, :to => 'self.class'
  end

  class Resource

    include ActiveResource::Hel::Dirty
    extend ActiveModel::Naming
  end
end
